"""
Thermal Noise Simulation in a Flow Cytometer Detector
=====================================================

This example demonstrates how thermal noise influences the raw signal in a flow cytometer detector.
We simulate thermal noise for detectors at two different temperatures and compare the resulting
signal characteristics.

Thermal noise, also known as Johnson-Nyquist noise, is generated by the thermal agitation of charge
carriers in the detector's resistive components. Its magnitude is proportional to the temperature
of the detector.

"""

from FlowCyPy.detector import Detector
from FlowCyPy.units import watt, ohm, ampere, second, hertz, degree, AU, kelvin
import matplotlib.pyplot as plt
from FlowCyPy import NoiseSetting

# Configure Noise Settings
NoiseSetting.include_noises = True
NoiseSetting.include_shot_noise = False
NoiseSetting.include_dark_current_noise = False
NoiseSetting.include_thermal_noise = True
NoiseSetting.include_RIN_noise = False

# Define simulation parameters
temperatures = [50, 500]  # Temperatures in Kelvin

# Initialize the plot
figure, (ax, ax1) = plt.subplots(2, 1, figsize=(10, 4))

# Simulate thermal noise for different temperatures
for temperature in temperatures:
    # Initialize Detector
    detector = Detector(
        name=str(temperature),
        responsitivity=1 * ampere / watt,
        resistance=50 * ohm,
        numerical_aperture=0.2 * AU,
        sampling_freq=1e6 * hertz,
        phi_angle=0 * degree,
        temperature=temperature * kelvin
    )
    detector.init_raw_signal(run_time=500e-6 * second)

    # Add thermal noise to the raw signal
    detector._add_thermal_noise_to_raw_signal()

    # Plot the raw signal on the top subplot
    detector.plot(ax=ax, add_raw=True, add_captured=False, show=False)

    # Plot the histogram of the raw signal on the bottom subplot
    ax1.hist(detector.dataframe['RawSignal'], alpha=0.5, label=f"{detector.name} K", bins=50)

# Add legends and show the plot
ax1.legend(title="Temperature")
ax.set_title("Raw Signal with Thermal Noise")
ax1.set_title("Histogram of Raw Signal")
plt.tight_layout()
plt.show()
